// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

#include "stdafx.h"
#include "mockvaluebase.h"

CMockValueBase::~CMockValueBase()
{
}
