// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

#define Base64_Encoder real_Base64_Encoder
#define Base64_Encode_Bytes real_Base64_Encode_Bytes
#define Base64_Decoder real_Base64_Decoder

#define GBALLOC_H

#include "base64.c"
