// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

#ifndef UAMQP_H
#define UAMQP_H

/* This header includes all the uAMQP headers
Unless advanced cherrypicking of which functionality is to be used,
a user can always inlcude this header */

#include "azure_uamqp_c/amqp_definitions.h"
#include "azure_uamqp_c/amqp_frame_codec.h"
#include "azure_uamqp_c/amqp_management.h"
#include "azure_uamqp_c/amqp_types.h"
#include "azure_uamqp_c/amqpvalue.h"
#include "azure_uamqp_c/amqpvalue_to_string.h"
#include "azure_uamqp_c/cbs.h"
#include "azure_uamqp_c/connection.h"
#include "azure_uamqp_c/frame_codec.h"
#include "azure_uamqp_c/header_detect_io.h"
#include "azure_uamqp_c/link.h"
#include "azure_uamqp_c/message.h"
#include "azure_uamqp_c/message_receiver.h"
#include "azure_uamqp_c/message_sender.h"
#include "azure_uamqp_c/messaging.h"
#include "azure_uamqp_c/sasl_anonymous.h"
#include "azure_uamqp_c/sasl_frame_codec.h"
#include "azure_uamqp_c/sasl_mechanism.h"
#include "azure_uamqp_c/sasl_server_mechanism.h"
#include "azure_uamqp_c/sasl_mssbcbs.h"
#include "azure_uamqp_c/sasl_plain.h"
#include "azure_uamqp_c/saslclientio.h"
#include "azure_uamqp_c/sasl_server_io.h"
#include "azure_uamqp_c/server_protocol_io.h"
#include "azure_uamqp_c/session.h"
#include "azure_uamqp_c/socket_listener.h"

#endif /* UAMQP_H */

